#include <queue>

#include <stdint.h>

#include <kernel/async/lock.hpp>
#include <kernel/log.hpp>
#include <kernel/mem/paging.hpp>
#include <kernel/mem/phys.hpp>
#include <kernel/signal.hpp>
#include <kernel/task/readyqueue.hpp>
#include <kernel/task/thread.hpp>

constexpr std::size_t KERNEL_STACK_ORDER = 3; // 2^3 * 4096 = 32KB

using namespace kernel::task;
using namespace kernel::mem::paging;

thread::thread(std::string name, pid_t owner)
    : owner { owner }, attr { READY | SYSTEM }, name { name } { }

thread::thread(const thread& val, pid_t owner)
    : owner { owner }, attr { val.attr }, name { val.name } { }

tid_t thread::tid() const
{
    return (tid_t)kstack.pfn;
}

bool thread::operator<(const thread& rhs) const
{
    return tid() < rhs.tid();
}

bool thread::operator==(const thread& rhs) const
{
    return tid() == rhs.tid();
}

static inline uintptr_t __stack_bottom(pfn_t pfn)
{
    return (uintptr_t)(void*)
        kernel::mem::physaddr<void>{pfn + (1 << KERNEL_STACK_ORDER) * 0x1000};
}

thread::kernel_stack::kernel_stack()
{
    pfn = page_to_pfn(alloc_pages(KERNEL_STACK_ORDER));
    sp = __stack_bottom(pfn);
}

thread::kernel_stack::kernel_stack(const kernel_stack& other)
    : kernel_stack()
{
    auto offset = __stack_bottom(other.pfn) - other.sp;

    sp -= offset;
    memcpy((void*)sp, (void*)other.sp, offset);
}

thread::kernel_stack::kernel_stack(kernel_stack&& other)
    : pfn(std::exchange(other.pfn, 0))
    , sp(std::exchange(other.sp, 0)) { }

thread::kernel_stack::~kernel_stack()
{
    if (!pfn)
        return;
    free_pages(pfn, KERNEL_STACK_ORDER);
}

uint64_t thread::kernel_stack::pushq(uint64_t val)
{
    *(uint64_t*)sp = val;
    sp -= 8;
    return val;
}

uint32_t thread::kernel_stack::pushl(uint32_t val)
{
    *(uint32_t*)sp = val;
    sp -= 4;
    return val;
}

void thread::set_attr(thd_attr_t new_attr)
{
    switch (new_attr) {
    case SYSTEM:
        attr |= SYSTEM;
        break;
    case READY:
        if (attr & ZOMBIE) {
            kmsgf("[kernel:warn] zombie process pid%d tries to wake up", owner);
            break;
        }

        if (attr & READY) {
            kmsgf("[kernel:warn] pid%d tries to wake up from ready state", owner);
            break;
        }

        attr &= SYSTEM;
        attr |= READY;

        dispatcher::enqueue(this);
        break;
    case ISLEEP:
        attr &= SYSTEM;
        attr |= ISLEEP;

        dispatcher::dequeue(this);
        break;
    case STOPPED:
        attr &= SYSTEM;
        attr |= STOPPED;

        dispatcher::dequeue(this);
        break;
    case ZOMBIE:
        attr &= SYSTEM;
        attr |= ZOMBIE;

        dispatcher::dequeue(this);
        break;
    default:
        kmsgf("[kernel:warn] unknown thread attribute: %x", new_attr);
        break;
    }
}

void thread::send_signal(signal_list::signo_type signal)
{
    if (signals.raise(signal))
        this->set_attr(READY);
}

int thread::set_thread_area(kernel::user::user_desc* ptr)
{
    if (ptr->read_exec_only && ptr->seg_not_present) {
        // TODO: use copy_to_user
        auto* dst = (void*)(uintptr_t)ptr->base_addr;
        std::size_t len = ptr->limit;
        if (len > 0 && dst)
            memset(dst, 0x00, len);
        return 0;
    }

    if (ptr->entry_number == -1U)
        ptr->entry_number = 6;
    else
        return -1;

    if (!ptr->seg_32bit)
        return -1;

    tls_desc[0]  = ptr->limit & 0x0000'ffff;
    tls_desc[0] |= (ptr->base_addr & 0x00ff'ffffULL) << 16;
    tls_desc[0] |= 0xe2'00'0000'0000;
    tls_desc[0] |= (ptr->limit & 0x000f'0000ULL) << (48-16);
    tls_desc[0] |= ((ptr->limit_in_pages + 0ULL) << 55);
    tls_desc[0] |= (ptr->base_addr & 0xf000'0000) << (56-28);

    tls_desc[1]  = 0; // 63:32: all 0, 31:0: ptr->base_addr[63:32]

    return 0;
}

int thread::load_thread_area() const
{
    kernel::user::load_thread_area(tls_desc[0], tls_desc[1]);
    return 0;
}
